/*
 * interval_set.c
 *
 *  Created on: Sep 20, 2010
 *      Author: parkcs
 *
 *  Implementation for efficient operations on a set of intervals
 *  Operations include O(log n) add/remove/search and O(n) full traversal
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "interval_set.h"

void* interval_set_allocator(unsigned int level)
{
	int size = sizeof(interval_set) + sizeof(void*) * level;
	void* ret = malloc(size);
	memset(ret, 0, size);
	return ret;
}

int interval_comparator(void* l, void* r)
{
	uintptr_t ls = ((addr_interval*)l)->start_addr;
	uintptr_t rs = ((addr_interval*)r)->start_addr;

	if (ls < rs)
		return -1;
	if (ls > rs)
		return 1;

	uintptr_t le = ((addr_interval*)l)->end_addr;
	uintptr_t re = ((addr_interval*)r)->end_addr;

	return (le < re) ? -1 : (le > re ? 1 : 0);
}

void interval_printer(addr_interval* i)
{
	if(i) {
		printf("%c%p,%p%c", is_left_closed(*i) ? '[' : '(',
				i->start_addr, i->end_addr,
				is_right_closed(*i) ? ']' : ')');
	}
}

void interval_set_printer(interval_set* s)
{
	printf("{ ");
	if(s) {
		skiplist_node* x = s->head;
		while(x) {
			if(x->data) {
				interval_printer((addr_interval*)x->key);
				x = x->forward[0];
				break;
			}
			x = x->forward[0];
		}
		while(x) {
			if(x->data) {
				printf(", ");
				interval_printer((addr_interval*)x->key);
			}
			x = x->forward[0];
		}
	}
	printf(" }");
}

interval_set* new_interval_set()
{
	interval_set* ret = (interval_set*) malloc(sizeof(interval_set));
	ret->allocator = interval_set_allocator;
	ret->comparator = interval_comparator;
	ret->prettyprinter = NULL;
	ret->level = 0;
	ret->head = ret->allocator(MAX_SKIPLIST_LEVEL - 1);
	return ret;
}

void delete_interval_set(interval_set* s)
{
	skiplist_node* x = s->head;
	while(x) {
		skiplist_node* tmp = x;
		x = x->forward[0];
		free(tmp);
	}
	free(s);

}

uintptr_t interval_set_hashcode(interval_set* s)
{
	uintptr_t hashcode = 0;
	skiplist_node* n = s->head;

	while(n) {
		if(n->data)
			hashcode += (uintptr_t)n->key;
		n = n->forward[0];
	}

	return hashcode;
}

bool interval_set_add(interval_set* s, addr_interval* i)
{
	skiplist_node* x = skiplist_insert(s, i, (void*)1);
	if(x) {
		if(x->data)
			return false;

		x->data = (void*)1;
	}
	return true;
}

bool interval_set_addall(interval_set* s, interval_set* is)
{
	bool ret = true;
	if(is) {
		skiplist_node* x = is->head;
		while(x) {
			if(x->data)
				ret = interval_set_add(s, x->key) && ret;
			x = x->forward[0];
		}
	}
	else
		ret = false;

	return ret;
}

bool interval_set_remove(interval_set* s, addr_interval* i)
{
	skiplist_node* x = skiplist_insert(s, i, (void*)0);
	if(x && x->data) {
		x->data = (void*)0;
		return true;
	}
	return false;
}

int interval_set_size(interval_set* s)
{
	if(s) {
		int count = 0;
		skiplist_node* x = s->head;
		while(x) {
			if(x->data)
				count++;
			x = x->forward[0];
		}
		return count;
	}
	return 0;
}

