/*
 * upct_racer.c -- online race detection for UPC
 *
 */

#ifndef UPCR_THRILLE_RACER
    #define UPCR_THRILLE_RACER  // Make sure all recursive calls into thrille are properly suffixed
#endif

#include "upct_types.h"
#include "upct_set.h" 
#include "upct_list.h"
#include "upct_multiset.h"

#include "upct_common.c"

int race_check(shared race_info_t *);

int lock_lt(shared void * l1, shared void * l2) {
    shared lock_addr * ll1 = (shared lock_addr *) l1;
    lock_addr addr1 = (lock_addr) (*ll1);
    shared lock_addr * ll2 = (shared lock_addr *) l2;
    lock_addr addr2 = (lock_addr) (*ll2);
    return addr1 < addr2;
}

void lock_print(shared void * lock) {
    shared lock_addr * my_lock = (shared lock_addr *) lock;
    lock_addr lock_id = (lock_addr) (*my_lock);
    printf("0x%lx", lock_id);
}

int race_info_lt(shared void * el1, shared void * el2) {
    shared race_info_t * left_ev = (shared race_info_t *) el1;
    shared race_info_t * right_ev = (shared race_info_t *) el2;

    return left_ev->addr < right_ev->addr;
}

void race_info_print(shared void * ev) {
    shared race_info_t * event = (shared race_info_t *) ev;
    print_race_info(event);
    printf(",\n");
}

/* Shared Data */
shared multiset_t * shared race_events;
shared int race_id;

/* Per Thread Data */
upc_lock_t * thrille_lock;
thrille_info_t thrille_info;

void thrille_init(void) {
    thrille_info.inside_inst = 0;
    THRILLE_BEGIN_FUNC();

    if (MYTHREAD == 0) {
        printf("Initializing Race Detection...\n");
        race_id = 0;
    }

    /* Global Allocations */
    thrille_lock = upc_all_lock_alloc();

    if (MYTHREAD == 0) {
        race_events = (shared multiset_t *) upc_alloc(sizeof(multiset_t));
        multiset_init(race_events, &race_info_lt);
    }

    /* Local Allocations */
    thrille_info.ls = (shared set_t *) upc_alloc(sizeof(set_t));
    set_init(thrille_info.ls, &lock_lt);
    thrille_info.phase = 0;

    upc_barrier;

    THRILLE_END_FUNC;
}

void thrille_before_read_racer(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
	    int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();
    THRILLE_LOCK;

    shared race_info_t * event;
    event = (shared race_info_t *) upc_alloc(sizeof(race_info_t));
    event->addr = mem_addr;
    event->thread = MYTHREAD;
    event->access = READ;
    event->locklist = lockset_getlist(thrille_info.ls);
    event->file = file;
    event->line = line;
    event->expr = expr;	
    event->race_reported = NOT_REPORTED;
    event->phase = thrille_info.phase;
    event->owned = (MYTHREAD == owner) ? 1 : 0;

#ifdef THRILLE_DEBUG
    printf("[%d] Before Read of Memory %p\n", MYTHREAD, mem_addr);
#endif
    if (NO_RACE == race_check(event)) {
        multiset_add(race_events, (shared void *) event);
    } else {
        upc_free(event);
    }
    /* multiset_print(race_events, &race_info_print); */

    THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_after_read_racer(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
	    int nBytes,
        int owner) { }

void thrille_before_write_racer(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
	    int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();
    THRILLE_LOCK;
    shared race_info_t * event;
    event = (shared race_info_t *) upc_alloc(sizeof(race_info_t));

    event->addr = mem_addr;
    event->thread = MYTHREAD;
    event->access = WRITE;
    event->locklist = lockset_getlist(thrille_info.ls);
    event->file = file;
    event->line = line;
    event->expr = expr;
    event->race_reported = NOT_REPORTED;
    event->phase = thrille_info.phase;
    event->owned = (MYTHREAD == owner) ? 1 : 0;


#ifdef THRILLE_DEBUG
    printf("[%d] Before Write of Memory 0x%p\n", MYTHREAD, mem_addr);
#endif
    if (NO_RACE == race_check(event)) {
        multiset_add(race_events, (shared void *) event);
    } else {
        upc_free(event);
    }
    /* multiset_print(race_events, &race_info_print); */

    THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_after_write_racer(int myid,
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
	    int nBytes,
        int owner) { }

void thrille_before_lock_racer(int myid, 
        uint64_t lock, 
        const char* file, 
        unsigned int line, 
        const char* expr) { }

void thrille_after_lock_racer(int myid, 
        uint64_t lock, 
        const char* file, 
        unsigned int line, 
        const char* expr) {
    THRILLE_BEGIN_FUNC();
    THRILLE_LOCK;

#ifdef THRILLE_DEBUG
    printf("[%d] After lock 0x%p\n", myid, lock);
#endif
    shared lock_addr * lock_data;
    lock_data = (shared lock_addr *) upc_alloc(sizeof(lock_addr));
    (*lock_data) = (lock_addr) lock;
    set_add(thrille_info.ls, (shared void *) lock_data);
    /* set_print(thrille_info.ls, &lock_print); */

    THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_before_unlock_racer(int myid, 
        uint64_t lock, 
        const char* file, 
        unsigned int line, 
        const char* expr) {
    THRILLE_BEGIN_FUNC();
    THRILLE_LOCK;

#ifdef THRILLE_DEBUG
    printf("[%d] Before unlock 0x%p\n", myid, lock);
#endif
    shared lock_addr * lock_data;
    lock_data = (shared lock_addr *) upc_alloc(sizeof(lock_addr));
    (*lock_data) = (lock_addr) lock;
    set_remove(thrille_info.ls, (shared void *) lock_data);
    /* set_print(thrille_info.ls, &lock_print); */

    THRILLE_UNLOCK;
    THRILLE_END_FUNC;
}

void thrille_after_unlock_racer(int myid,
        uint64_t lock, 
        const char* file, 
        unsigned int line, 
        const char* expr) { }

void thrille_before_notify_racer(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { }
void thrille_after_notify_racer(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { 
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
#ifdef THRILLE_DEBUG
    printf("[%d] Now in phase %d (After notify)\n", myid, thrille_info.phase);
#endif
    THRILLE_END_FUNC;
}
void thrille_before_wait_racer(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { }
void thrille_after_wait_racer(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { 
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
#ifdef THRILLE_DEBUG
    printf("[%d] Now in phase %d (After wait)\n", myid, thrille_info.phase);
#endif
    THRILLE_END_FUNC;
}
void thrille_before_try_wait_racer(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { }
void thrille_after_try_wait_racer(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { 
    /*THRILLE_BEGIN_FUNC(ret_val);
    if(ret_val) {
        thrille_info.phase++;
        //printf("[%d] Now in phase %d (After successful try-wait)\n", myid, thrille_info.phase);
    }
    THRILLE_END_FUNC;
    return ret_val;*/
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //printf("[%d] Now in phase %d (After try_wait)\n", myid, thrille_info.phase);
    THRILLE_END_FUNC;
}


int is_racing(shared race_info_t * ev1, shared race_info_t * ev2) {
    if (ev1->addr != ev2->addr) {
        return NO_RACE;
    }
    if ((ev1->access == READ) && (ev2->access == READ)) {
        return NO_RACE;
    }
    if (ev1->thread == ev2->thread) {
        return NO_RACE;
    }
    if (list_intersection(ev1->locklist, ev2->locklist, &lock_lt) == NONEMPTY){
        return NO_RACE;
    }
    if (ev1->phase % 2) { // ODD phase -- after nofity (race with +-2)
        if((ev2->phase - ev1->phase < -2) || (ev2->phase - ev1->phase > 2))
            return NO_RACE;
    }
    else { // EVEN phase -- after wait (race with +-1)
        if((ev2->phase - ev1->phase < -1) || (ev2->phase - ev1->phase > 1))
            return NO_RACE;
    }
    return RACE;
}


void report_race(shared race_info_t * ev1, shared race_info_t * ev2) {
    ++race_id;
    printf("POTENTIAL RACE #%d FOUND BETWEEN:\n", race_id);
    printf("\t");
    print_race_info(ev1);
    printf("\n\t");
    print_race_info(ev2);
    printf("\n");
    write_race_pair(ev1, ev2, race_id);
}

int race_check(shared race_info_t * event) {
    shared list_elem * current = multiset_get(race_events, event);
    shared race_info_t * other = NULL;
    while (current != NULL) {
        other = (shared race_info_t *) list_get_data(current);
        if (other->race_reported == REPORTED) {
            current = list_get_next(current);
            continue;
        }
        if (is_racing(event, other) == RACE) {
            other->race_reported = REPORTED;
            event->race_reported = REPORTED;
            report_race(event, other);
            return RACE;
        }
        current = list_get_next(current);
    }
    return NO_RACE;
}
