/*
 * upc_tester.c -- active testing for upc
 *
 */

#ifndef UPCR_THRILLE_TESTER
    #define UPCR_THRILLE_TESTER  // Make sure all recursive calls into thrille are properly suffixed
#endif
#include "upct_types.h"

#include "upct_common.c"

#define UPCT_RACE_ID_ENV_STRING	"UPCT_RACE_ID"
#define UPCT_RACE_RESOLVE_ENV_STRING	"UPCT_RACE_RESOLVE_DIR"

struct _race_target_t {
    const char* file;
    int line;
};


/* Shared Data */
shared ptr_sem *semaphores;
shared race_info_t *races;
shared int num_thrash = 0;
shared int num_reproduced = 0;

/* Per Thread Data */
upc_lock_t *thrille_lock;
thrille_info_t thrille_info;
race_info_t target1, target2;
int race_resolve_dir;

void thrille_init(void) {
	thrille_info.inside_inst = 0;
	THRILLE_BEGIN_FUNC();
    if (MYTHREAD == 0) {
        printf("UPC Active Testing started...\n");
    }

	char* race_id_str = bupc_getenv(UPCT_RACE_ID_ENV_STRING);
	int id = race_id_str ? atoi(race_id_str) : 0;
    char* race_resolve_str = bupc_getenv(UPCT_RACE_RESOLVE_ENV_STRING);
    race_resolve_dir = race_resolve_str ? atoi(race_resolve_str) : 0; 
    if (MYTHREAD == 0)
		printf("Reading race pair %d\n", id);

	read_race_pair(&target1, &target2, id);
 
	if (MYTHREAD == 0) {
		print_race_info_ns(&target1);
		printf("\n===\n");
		print_race_info_ns(&target2);
		printf("\n");
        printf("Resolve order: %d\n", race_resolve_dir);
	}

    //setRaceTargets();

    thrille_lock = upc_all_lock_alloc();	
    semaphores = (shared ptr_sem*) upc_all_alloc(THREADS, sizeof(ptr_sem));
    semaphores[MYTHREAD] = bupc_sem_alloc(0);

    races = (shared race_info_t*) upc_all_alloc(THREADS, sizeof(race_info_t));
    races[MYTHREAD].addr = 0;
    thrille_info.phase = 0;

    upc_barrier;
	THRILLE_END_FUNC;
}

int isRacingPair(const char* file, unsigned int line, int owned) {
    //printf("%s %s\n", file, target1.file);
    int isRacingPair1 = (strcmp(file, target1.file) == 0) && 
        (line == target1.line) && 
        (owned == target1.owned) &&
        (thrille_info.phase == target1.phase);
    
    int isRacingPair2 = (strcmp(file, target2.file) == 0) && 
        (line == target2.line) &&
        (owned == target2.owned) &&
        (thrille_info.phase == target2.phase);
    
    return isRacingPair1 + 2*isRacingPair2;
}

int checkPending(race_info_t event) {
    for (int racingThread = 0; racingThread < THREADS; racingThread++) {
        if (racingThread == event.thread) {
            continue;
        }
        race_info_t other = races[racingThread];
        int sameAddr = (other.addr == event.addr);
        int oneWrite = (event.access == WRITE || 
                other.access == WRITE);
        int sameTarget = (event.race_reported == 1 && other.race_reported == 1)
                || (event.race_reported == 2 && other.race_reported == 2);
        if (sameAddr) {
            if (oneWrite && !sameTarget)
            //if (sameAddr && oneWrite && !sameTarget)
                return racingThread;
            else
                return BAD_RACE;
        }
	else if (sameTarget)
		return BAD_RACE;
    }
    return NO_RACE;
}

void raceFound(race_info_t ev1, race_info_t ev2) {
    printf("REAL RACE CONFIRMED FOR ADDRESS 0x%lx\n", ev1.addr);
    printf("Racing pairs:\n");
    printf("\t(T%d:%s@%d:%s)\n", ev1.thread, ev1.file, ev1.line, ev1.expr);
    printf("\t(T%d:%s@%d:%s)\n", ev2.thread, ev2.file, ev2.line, ev2.expr);
}

void registerWaiting(race_info_t event) {
    races[MYTHREAD] = event;
}

void unregisterWaiting(race_info_t event) {
    races[MYTHREAD].addr = 0;
}

void printRaceInfoT(race_info_t event) {
    printf("race_info_t(");
    printf("Thread: %d, ", event.thread);
    printf("Addr: 0x%lx, ", event.addr);
    printf("Access Type: %s, ", event.access == WRITE ? "W" : "R");
    printf("File: %s, ", event.file);
    printf("Line: %d, ", event.line);
    printf("Expr: %s, ", event.expr);
    printf(")\n");
}

void randomActive(race_info_t event) {
    THRILLE_LOCK;
    int racingThread = checkPending(event);
    if (racingThread == NO_RACE) {
        registerWaiting(event);
        int n = 0;
        printf("[%d] paused at line %d to try to create a real race (Addr: 0x%lx%s)\n"
                , MYTHREAD, event.line, event.addr, event.access == WRITE ? "W" : "R");
        int realRace = 0;
        while(!(realRace = bupc_sem_try(semaphores[MYTHREAD])) && 
                n++ < MAX_SEM_TRIES) {
            THRILLE_UNLOCK;
            usleep(SEM_WAIT_USEC);
            THRILLE_LOCK;
        }
        printf("[%d] is no longer paused\n", MYTHREAD);
        unregisterWaiting(event);
        if(realRace) {
            ++num_reproduced;
            printf("%d reproduced (%d thrashes)\n", num_reproduced, num_thrash);
            if ((race_resolve_dir && event.race_reported == 1) ||
                    (!race_resolve_dir && event.race_reported == 2)) {
                THRILLE_UNLOCK;
                //printf("[%d] Waiting guy sleeping...(dir = %d, target = %d)\n", MYTHREAD, race_resolve_dir, event.race_reported);
                usleep(10*SEM_WAIT_USEC);
                THRILLE_LOCK;
            }
        }
        else {
            ++num_thrash;
            printf("%d thrashes (%d reproduced)\n", num_thrash, num_reproduced);
        }
    } else if (racingThread != BAD_RACE) {
        raceFound(event, races[racingThread]);
        bupc_sem_post(semaphores[racingThread]);
        if ((race_resolve_dir && event.race_reported == 1) ||
                (!race_resolve_dir && event.race_reported == 2)) {
            THRILLE_UNLOCK;
            //printf("[%d] Posting guy sleeping...(dir = %d, target = %d)\n", MYTHREAD, race_resolve_dir, event.race_reported);
            usleep(10*SEM_WAIT_USEC);
            THRILLE_LOCK;
        }
    }
    THRILLE_UNLOCK;
}

void thrille_before_read_tester(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();
    int r;
    int owned = (MYTHREAD == owner) ? 1: 0;
    if (r = isRacingPair(file, line, owned)) {
        race_info_t raceEvent;
        raceEvent.addr = mem_addr;
        raceEvent.thread = myid;
        raceEvent.access = READ;
        raceEvent.file = file;
		//ls_strcpy(file, raceEvent.file);
        raceEvent.line = line;
        raceEvent.expr = expr;
        raceEvent.phase = thrille_info.phase;
		//ls_strcpy(expr, raceEvent.expr);
        raceEvent.race_reported = r;
        raceEvent.owned = owned;
        randomActive(raceEvent);
    }
    THRILLE_END_FUNC;
}

void thrille_after_read_tester(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
	    int nBytes,
        int owner) { }

void thrille_before_write_tester(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();
    int r;
    int owned = (MYTHREAD == owner) ? 1: 0;
    if(r = isRacingPair(file, line, owned)) {
        race_info_t raceEvent;
        raceEvent.thread = myid;
        raceEvent.addr = mem_addr;
        raceEvent.access = WRITE;
        raceEvent.file = file;
		//ls_strcpy(file, raceEvent.file);
        raceEvent.line = line;
        raceEvent.expr = expr;
        raceEvent.phase = thrille_info.phase;
		//ls_strcpy(expr, raceEvent.expr);
        raceEvent.race_reported = r;
        raceEvent.owned = owned;
        randomActive(raceEvent);
    }
    THRILLE_END_FUNC;
}

void thrille_after_write_tester(int myid, 
        uint64_t mem_addr, 
        const char* file, 
        unsigned int line, 
        const char* expr,
        int nBytes,
        int owner) { }

void thrille_before_lock_tester(int myid,
        uint64_t lock, 
        const char* file, 
        unsigned int line, 
        const char* expr) { }

void thrille_after_lock_tester(int myid, 
                        uint64_t lock_addr, 
                        const char* file, 
                        unsigned int line, 
                        const char* expr) {
    THRILLE_BEGIN_FUNC();
    //printf("[%d] After lock 0x%lx\n", myid, lock_addr);
    THRILLE_END_FUNC;
}

void thrille_before_unlock_tester(int myid, 
                          uint64_t lock_addr, 
                          const char* file, 
                          unsigned int line, 
                          const char* expr) {
    THRILLE_BEGIN_FUNC();
    //printf("[%d] Before unlock 0x%lx\n", myid, lock_addr);
    THRILLE_END_FUNC;
}

void thrille_after_unlock_tester(int myid,
        uint64_t lock, 
        const char* file, 
        unsigned int line, 
        const char* expr) { }

void thrille_before_notify_tester(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { }

void thrille_after_notify_tester(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { 
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //printf("[%d] Now in phase %d (After notify)\n", myid, thrille_info.phase);
    THRILLE_END_FUNC;
}

void thrille_before_wait_tester(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { }

void thrille_after_wait_tester(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { 
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //printf("[%d] Now in phase %d (After wait)\n", myid, thrille_info.phase);
    THRILLE_END_FUNC;
}

void thrille_before_try_wait_tester(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { }

void thrille_after_try_wait_tester(int myid, int barrier_val, const char* file, unsigned int line, const char* expr) { 
    /*THRILLE_BEGIN_FUNC(ret_val);
    if(ret_val) {
        thrille_info.phase++;
        //printf("[%d] Now in phase %d (After successful try-wait)\n", myid, thrille_info.phase);
    }
    THRILLE_END_FUNC;
    return ret_val;*/
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //printf("[%d] Now in phase %d (After try_wait)\n", myid, thrille_info.phase);
    THRILLE_END_FUNC;
}

